;
; -------------------------------
; THE 'PREPARE TO ADD' SUBROUTINE
; -------------------------------
; This routine is called twice to prepare each floating point number for
; addition, in situ, on the calculator stack.
; The exponent is picked up from the first byte which is then cleared to act
; as a sign byte and accept any overflow.
; If the exponent is zero then the number is zero and an early return is made.
; The now redundant sign bit of the mantissa is set and if the number is 
; negative then all five bytes of the number are twos-complemented to prepare 
; the number for addition.
; On the second invocation the exponent of the first number is in B.
;
;; PREP-ADD
L16D8:	LD A,(HL)		; fetch exponent.
	LD (HL),$00		; make this byte zero to take any overflow and
				; default to positive.
	AND A			; test stored exponent for zero.
	RET Z			; return with zero flag set if number is zero.

	INC HL			; point to first byte of mantissa.
	BIT 7,(HL)		; test the sign bit.
	SET 7,(HL)		; set it to its implied state.
	DEC HL			; set pointer to first byte again.
	RET Z			; return if bit indicated number is positive.>>
;
; if negative then all five bytes are twos complemented starting at LSB.
;
	PUSH BC			; save B register contents.
	LD BC,$0005		; set BC to five.
	ADD HL,BC		; point to location after 5th byte.
	LD B,C			; set the B counter to five.
	LD C,A			; store original exponent in C.
	SCF			; set carry flag so that one is added.
;
; now enter a loop to twos-complement the number.
; The first of the five bytes becomes $FF to denote a negative number.
;
;; NEG-BYTE
L16EC:	DEC HL			; point to first or more significant byte.
	LD A,(HL)		; fetch to accumulator.
	CPL			; complement.
	ADC A,$00		; add in initial carry or any subsequent carry.
	LD (HL),A		; place number back.
	DJNZ L16EC		; loop back five times to NEG-BYTE

	LD A,C			; restore the exponent to accumulator.
	POP BC			; restore B register contents.

	RET			; return.
;
; ----------------------------------
; THE 'FETCH TWO NUMBERS' SUBROUTINE
; ----------------------------------
; This routine is used by addition, multiplication and division to fetch
; the two five-byte numbers addressed by HL and DE from the calculator stack
; into the Z80 registers.
; The HL register may no longer point to the first of the two numbers.
; Since the 32-bit addition operation is accomplished using two Z80 16-bit
; instructions, it is important that the lower two bytes of each mantissa are
; in one set of registers and the other bytes all in the alternate set.
;
; In: HL = highest number, DE= lowest number
;
;	: alt'	:	:
; Out:	:H,B-C	:C,B	: num1
;	:L,D-E	:D-E	: num2
;
;; FETCH-TWO
L16F7:	PUSH HL	; save HL 
	PUSH AF	; save A - result sign when used from division.

	LD C,(HL)		;
	INC HL			;
	LD B,(HL)		;
	LD (HL),A		; insert sign when used from multiplication.
	INC HL			;
	LD A,C			; m1
	LD C,(HL)		;
	PUSH BC			; PUSH m2 m3

	INC HL			;
	LD C,(HL)		; m4
	INC HL			;
	LD B,(HL)		; m5  BC holds m5 m4

	EX DE,HL		; make HL point to start of second number.

	LD D,A			; m1
	LD E,(HL)		;
	PUSH DE			; PUSH m1 n1

	INC HL			;
	LD D,(HL)		;
	INC HL			;
	LD E,(HL)		;
	PUSH DE			; PUSH n2 n3

	EXX			; - - - - - - -

	POP DE			; POP n2 n3
	POP HL			; POP m1 n1
	POP BC			; POP m2 m3

	EXX			; - - - - - - -

	INC HL			;
	LD D,(HL)		;
	INC HL			;
	LD E,(HL)		; DE holds n4 n5

	POP AF			; restore saved
	POP HL			; registers.
	RET 			; return.
;
; -----------------------------
; THE 'SHIFT ADDEND' SUBROUTINE
; -----------------------------
; The accumulator A contains the difference between the two exponents.
; This is the lowest of the two numbers to be added 
;
;; SHIFT-FP
L171A:	AND A			; test difference between exponents.
	RET Z			; return if zero. both normal.

	CP $21			; compare with 33 bits.
	JR NC,L1736		; forward if greater than 32 to ADDEND-0

	PUSH BC			; preserve BC - part 
	LD B,A			; shift counter to B.
;
; Now perform B right shifts on the addend  L'D'E'D E
; to bring it into line with the augend	H'B'C'C B
;
;; ONE-SHIFT
L1722:	EXX			; - - -
	SRA L			;    76543210->C    bit 7 unchanged.
	RR D			; C->76543210->C
	RR E			; C->76543210->C
	EXX			; - - - 
	RR D			; C->76543210->C
	RR E			; C->76543210->C
	DJNZ L1722		; loop back B times to ONE-SHIFT

	POP BC			; restore BC
	RET NC			; return if last shift produced no carry.   >>
;
; if carry flag was set then accuracy is being lost so round up the addend.
;
	CALL L1741		; routine ADD-BACK
	RET NZ			; return if not FF 00 00 00 00
;
; this branch makes all five bytes of the addend zero and is made during
; addition when the exponents are too far apart for the addend bits to 
; affect the result.
;
;; ADDEND-0
L1736:	EXX			; select alternate set for more significant 
				; bytes.
	XOR A			; clear accumulator.
;
; this entry point (from multiplication) sets four of the bytes to zero or if 
; continuing from above, during addition, then all five bytes are set to zero.
;
;; ZEROS-4/5
L1738:	LD L,$00		; set byte 1 to zero.
	LD D,A			; set byte 2 to A.
	LD E,L			; set byte 3 to zero.
	EXX			; select main set 
	LD DE,$0000		; set lower bytes 4 and 5 to zero.
	RET			; return.
;
; -------------------------
; THE 'ADD-BACK' SUBROUTINE
; -------------------------
; Called from SHIFT-FP above during addition and after normalization from
; multiplication.
; This is really a 32-bit increment routine which sets the zero flag according
; to the 32-bit result.
; During addition, only negative numbers like FF FF FF FF FF,
; the twos-complement version of xx 80 00 00 01 say 
; will result in a full ripple FF 00 00 00 00.
; FF FF FF FF FF when shifted right is unchanged by SHIFT-FP but sets the 
; carry invoking this routine.
;
;; ADD-BACK
L1741:	INC E			;
	RET NZ			;

	INC D			;
	RET NZ			;

	EXX			;
	INC E			;
	JR NZ,L174A		; forward if no overflow to ALL-ADDED

	INC D			;
;
;; ALL-ADDED
L174A:	EXX			;
	RET 			; return with zero flag set for zero mantissa.
;
; ---------------------------
; THE 'SUBTRACTION' OPERATION
; ---------------------------
; just switch the sign of subtrahend and do an add.
;
;; subtract
L174C:	LD A,(DE)		; fetch exponent byte of second number the
				; subtrahend. 
	AND A			; test for zero
	RET Z			; return if zero - first number is result.

	INC DE			; address the first mantissa byte.
	LD A,(DE)		; fetch to accumulator.
	XOR $80			; toggle the sign bit.
	LD (DE),A		; place back on calculator stack.
	DEC DE			; point to exponent byte.
				; continue into addition routine.
;
; ------------------------
; THE 'ADDITION' OPERATION
; ------------------------
; The addition operation pulls out all the stops and uses most of the Z80's
; registers to add two floating-point numbers.
; This is a binary operation and on entry, HL points to the first number
; and DE to the second.
;
;; addition
L1755:	EXX			; - - -
	PUSH HL			; save the pointer to the next literal.
	EXX			; - - -

	PUSH DE			; save pointer to second number
	PUSH HL			; save pointer to first number - will be the
				; result pointer on calculator stack.

	CALL L16D8		; routine PREP-ADD
	LD B,A			; save first exponent byte in B.
	EX DE,HL		; switch number pointers.
	CALL L16D8		; routine PREP-ADD
	LD C,A			; save second exponent byte in C.
	CP B			; compare the exponent bytes.
	JR NC,L1769		; forward if second higher to SHIFT-LEN

	LD A,B			; else higher exponent to A
	LD B,C			; lower exponent to B
	EX DE,HL		; switch the number pointers.
;
;; SHIFT-LEN
L1769:	PUSH AF			; save higher exponent
	SUB B			; subtract lower exponent

	CALL L16F7		; routine FETCH-TWO
	CALL L171A		; routine SHIFT-FP

	POP AF			; restore higher exponent.
	POP HL			; restore result pointer.
	LD (HL),A		; insert exponent byte.
	PUSH HL			; save result pointer again.
;
; now perform the 32-bit addition using two 16-bit Z80 add instructions.
;
	LD L,B			; transfer low bytes of mantissa individually
	LD H,C			; to HL register

	ADD HL,DE		; the actual binary addition of lower bytes
;
; now the two higher byte pairs that are in the alternate register sets.
;
	EXX			; switch in set 
	EX DE,HL		; transfer high mantissa bytes to HL register.

	ADC HL,BC		; the actual addition of higher bytes with
				; any carry from first stage.

	EX DE,HL		; result in DE, sign bytes ($FF or $00) to HL
;
; now consider the two sign bytes
;
	LD A,H			; fetch sign byte of num1

	ADC A,L			; add including any carry from mantissa 
				; addition. 00 or 01 or FE or FF

	LD L,A			; result in L.
;
; possible outcomes of signs and overflow from mantissa are
;
;  H +  L + carry =  L    RRA  XOR L  RRA
; ------------------------------------------------------------
; 00 + 00	  = 00    00   00
; 00 + 00 + carry = 01    00   01	carry
; FF + FF	  = FE C  FF   01	carry
; FF + FF + carry = FF C  FF   00
; FF + 00	  = FF    FF   00
; FF + 00 + carry = 00 C  80   80
;
	RRA			; C->76543210->C
	XOR L			; set bit 0 if shifting required.

	EXX			; switch back to main set
	EX DE,HL		; full mantissa result now in D'E'D E registers.
	POP HL			; restore pointer to result exponent on 
				; the calculator stack.

	RRA			; has overflow occurred ?
	JR NC,L1790		; skip forward if not to TEST-NEG
;
; if the addition of two positive mantissas produced overflow or if the
; addition of two negative mantissas did not then the result exponent has to
; be incremented and the mantissa shifted one place to the right.
;
	LD A,$01		; one shift required.
	CALL L171A		; routine SHIFT-FP performs a single shift 
				; rounding any lost bit
	INC (HL)		; increment the exponent.
	JR Z,L17B3		; forward to ADD-REP-6 if the exponent
				; wraps round from FF to zero as number is too
				; big for the system.
;
; at this stage the exponent on the calculator stack is correct.
;
;; TEST-NEG
L1790:	EXX			; switch in the alternate set.
	LD A,L			; load result sign to accumulator.
	AND $80			; isolate bit 7 from sign byte setting zero
				; flag if positive.
	EXX			; back to main set.

	INC HL			; point to first byte of mantissa
	LD (HL),A		; insert $00 positive or $80 negative at 
				; position on calculator stack.

	DEC HL			; point to exponent again.
	JR Z,L17B9		; forward if positive to GO-NC-MLT
;
; a negative number has to be twos-complemented before being placed on stack.
;
	LD A,E			; fetch lowest (rightmost) mantissa byte.
	NEG			; Negate
	CCF			; Complement Carry Flag
	LD E,A			; place back in register

	LD A,D			; ditto
	CPL			;
	ADC A,$00		;
	LD D,A			;

	EXX			; switch to higher (leftmost) 16 bits.

	LD A,E			; ditto
	CPL			;
	ADC A,$00		;
	LD E,A			;

	LD A,D			; ditto
	CPL			;
	ADC A,$00		;
	JR NC,L17B7		; forward without overflow to END-COMPL
;
; else entire mantissa is now zero.  00 00 00 00
;
	RRA			; set mantissa to 80 00 00 00
	EXX			; switch.
	INC (HL)		; increment the exponent.
;
;; ADD-REP-6
L17B3:	JP Z,L1880		; jump forward if exponent now zero to REPORT-6
				; 'Number too big'

	EXX			; switch back to alternate set.
;
;; END-COMPL
L17B7:	LD D,A			; put first byte of mantissa back in DE.
	EXX			; switch to main set.

;; GO-NC-MLT
L17B9:	XOR A			; clear carry flag and
				; clear accumulator so no extra bits carried
				; forward as occurs in multiplication.

	JR L1828		; forward to common code at TEST-NORM 
				; but should go straight to NORMALIZE.
;
; ----------------------------------------------
; THE 'PREPARE TO MULTIPLY OR DIVIDE' SUBROUTINE
; ----------------------------------------------
; this routine is called twice from multiplication and twice from division
; to prepare each of the two numbers for the operation.
; Initially the accumulator holds zero and after the second invocation bit 7
; of the accumulator will be the sign bit of the result.
;
;; PREP-M/D
L17BC:	SCF			; set carry flag to signal number is zero.
	DEC (HL)		; test exponent
	INC (HL)		; for zero.
	RET Z			; return if zero with carry flag set.

	INC HL			; address first mantissa byte.
	XOR (HL)		; exclusive or the running sign bit.
	SET 7,(HL)		; set the implied bit.
	DEC HL			; point to exponent byte.
	RET 			; return.
;
; ------------------------------
; THE 'MULTIPLICATION' OPERATION
; ------------------------------
;
;; multiply
L17C6:	XOR A			; reset bit 7 of running sign flag.
	CALL L17BC		; routine PREP-M/D
	RET C			; return if number is zero.
				; zero * anything = zero.

	EXX			; - - -
	PUSH HL			; save pointer to 'next literal'
	EXX			; - - -

	PUSH DE			; save pointer to second number 

	EX DE,HL		; make HL address second number.

	CALL L17BC		; routine PREP-M/D

	EX DE,HL		; HL first number, DE - second number
	JR C,L1830		 ; forward with carry to ZERO-RSLT
				; anything * zero = zero.

	PUSH HL			; save pointer to first number.

	CALL L16F7		; routine FETCH-TWO fetches two mantissas from
				; calc stack to B'C'C,B  D'E'D E
				; (HL will be overwritten but the result sign
				; in A is inserted on the calculator stack)

	LD A,B			; transfer low mantissa byte of first number
	AND A			; clear carry.
	SBC HL,HL		; a short form of LD HL,$0000 to take lower
				; two bytes of result. (2 program bytes)
	EXX			; switch in alternate set
	PUSH HL			; preserve HL
	SBC HL,HL		; set HL to zero also to take higher two bytes
				; of the result and clear carry.
	EXX			; switch back.

	LD B,$21		; register B can now be used to count thirty 
				; three shifts.
	JR L17F8		; forward to loop entry point STRT-MLT
; ---
;
; The multiplication loop is entered at  STRT-LOOP.
;
;; MLT-LOOP
L17E7:	JR NC,L17EE		; forward if no carry to NO-ADD
				; else add in the multiplicand.

	ADD HL,DE		; add the two low bytes to result
	EXX			; switch to more significant bytes.
	ADC HL,DE		; add high bytes of multiplicand and any carry.
	EXX			; switch to main set.
;
; in either case shift result right into B'C'C A
;
;; NO-ADD
L17EE:	EXX			; switch to alternate set
	RR H			; C > 76543210 > C
	RR L			; C > 76543210 > C
	EXX			;
	RR H			; C > 76543210 > C
	RR L			; C > 76543210 > C
;
;; STRT-MLT
L17F8:	EXX			; switch in alternate set.
	RR B			; C > 76543210 > C
	RR C			; C > 76543210 > C
	EXX			; now main set
	RR C			; C > 76543210 > C
	RRA			; C > 76543210 > C
	DJNZ L17E7		; loop back 33 times to MLT-LOOP

	EX DE,HL		;
	EXX			;
	EX DE,HL		;
	EXX			;
	POP BC			;
	POP HL			;
	LD A,B			;
	ADD A,C			;
	JR NZ,L180E		; forward to MAKE-EXPT

	AND A			;
;
;; MAKE-EXPT
L180E:	DEC A			;
	CCF			; Complement Carry Flag
;
;; DIVN-EXPT
L1810:	RLA			;
	CCF			; Complement Carry Flag
	RRA			;
	JP P,L1819		; forward to OFLW1-CLR

	JR NC,L1880		; forward to REPORT-6

	AND A			;
;
;; OFLW1-CLR
L1819:	INC A			;
	JR NZ,L1824		; forward to OFLW2-CLR

	JR C,L1824		; forward to OFLW2-CLR

	EXX			;
	BIT 7,D			;
	EXX			;
	JR NZ,L1880		; forward to REPORT-6
;
;; OFLW2-CLR
L1824:	LD (HL),A		;
	EXX			;
	LD A,B			;
	EXX			;
;
; addition joins here with carry flag clear.
;
;; TEST-NORM
L1828:	JR NC,L183F		; forward to NORMALIZE

	LD A,(HL)		;
	AND A			;
;
;; NEAR-ZERO
L182C:	LD A,$80		; prepare to rescue the most significant bit 
				; of the mantissa if it is set.
	JR Z,L1831		; skip forward to SKIP-ZERO
;
;; ZERO-RSLT
L1830:	XOR A			; make mask byte zero signaling set five
				; bytes to zero.
;
;; SKIP-ZERO
L1831:	EXX			; switch in alternate set
	AND D			; isolate most significant bit (if A is $80).

	CALL L1738		; routine ZEROS-4/5 sets mantissa without 
				; affecting any flags.

	RLCA			; test if MSB set. bit 7 goes to bit 0.
				; either $00 -> $00 or $80 -> $01
	LD (HL),A		; make exponent $01 (lowest) or $00 zero
	JR C,L1868		; forward if first case to OFLOW-CLR

	INC HL			; address first mantissa byte on the
				; calculator stack.
	LD (HL),A		; insert a zero for the sign bit.
	DEC HL			; point to zero exponent
	JR L1868		; forward to OFLOW-CLR
; ---
;
; this branch is common to addition and multiplication with the mantissa
; result still in registers D'E'D E .
;
;; NORMALIZE
L183F:	LD B,$20		; a maximum of thirty-two left shifts will be 
				; needed.
;
;; SHIFT-ONE
L1841:	EXX			; address higher 16 bits.
	BIT 7,D			; test the leftmost bit
	EXX			; address lower 16 bits.

	JR NZ,L1859		; forward if leftmost bit was set to NORML-NOW

	RLCA			; this holds zero from addition, 33rd bit 
				; from multiplication.

	RL E			; C < 76543210 < C
	RL D			; C < 76543210 < C

	EXX			; address higher 16 bits.

	RL E			; C < 76543210 < C
	RL D			; C < 76543210 < C

	EXX			; switch to main set.

	DEC (HL)		; decrement the exponent byte on the calculator
				; stack.

	JR Z,L182C		; back if exponent becomes zero to NEAR-ZERO
				; it's just possible that the last rotation
				; set bit 7 of D. We shall see.

	DJNZ L1841		; loop back to SHIFT-ONE
;
; if thirty-two left shifts were performed without setting the most significant 
; bit then the result is zero.
;
	JR L1830	; back to ZERO-RSLT
; ---
;
;; NORML-NOW
L1859:	RLA			; for the addition path, A is always zero.
				; for the mult path, ...

	JR NC,L1868		; forward to OFLOW-CLR
;
; this branch is taken only with multiplication.
;
	CALL L1741		; routine ADD-BACK

	JR NZ,L1868		; forward to OFLOW-CLR

	EXX			;
	LD D,$80		;
	EXX			;
	INC (HL)		;
	JR Z,L1880		; forward to REPORT-6
;
; now transfer the mantissa from the register sets to the calculator stack
; incorporating the sign bit already there.
;
;; OFLOW-CLR
L1868:	PUSH HL			; save pointer to exponent on stack.
	INC HL			; address first byte of mantissa which was 
				; previously loaded with sign bit $00 or $80.

	EXX			; - - -
	PUSH DE			; push the most significant two bytes.
	EXX			; - - -

	POP BC			; pop - true mantissa is now BCDE.
;
; now pick up the sign bit.
;
	LD A,B			; first mantissa byte to A 
	RLA			; rotate out bit 7 which is set
	RL (HL)			; rotate sign bit on stack into carry.
	RRA			; rotate sign bit into bit 7 of mantissa.
;
; and transfer mantissa from main registers to calculator stack.
;
	LD (HL),A		;
	INC HL			;
	LD (HL),C		;
	INC HL			;
	LD (HL),D		;
	INC HL			;
	LD (HL),E		;

	POP HL			; restore pointer to num1 now result.
	POP DE			; restore pointer to num2 now STKEND.

	EXX			; - - -
	POP HL			; restore pointer to next calculator literal.
	EXX			; - - -

	RET			; return.
; ---
;
;; REPORT-6
L1880:	RST 08H			; ERROR-1
	.db $05			; Error Report: Arithmetic overflow.
;
; ------------------------
; THE 'DIVISION' OPERATION
; ------------------------
;   "Of all the arithmetic subroutines, division is the most complicated and
;   the least understood.  It is particularly interesting to note that the 
;   Sinclair programmer himself has made a mistake in his programming ( or has
;   copied over someone else's mistake!) for
;   PRINT PEEK 6352 [ $18D0 ] ('unimproved' ROM, 6351 [ $18CF ] )
;   should give 218 not 225."
;   - Dr. Ian Logan, Syntax magazine Jul/Aug 1982.
;   [  i.e. the jump should be made to div-34th ]
;
;   First check for division by zero.
;
;; division
L1882:	EX DE,HL		; consider the second number first. 
	XOR A			; set the running sign flag.
	CALL L17BC		; routine PREP-M/D
	JR C,L1880		; back if zero to REPORT-6
				; 'Arithmetic overflow'

	EX DE,HL		; now prepare first number and check for zero.
	CALL L17BC		; routine PREP-M/D
	RET C			; return if zero, 0/anything is zero.

	EXX			; - - -
	PUSH HL			; save pointer to the next calculator literal.
	EXX			; - - -

	PUSH DE			; save pointer to divisor - will be STKEND.
	PUSH HL			; save pointer to dividend - will be result.

	CALL L16F7		; routine FETCH-TWO fetches the two numbers
				; into the registers H'B'C'C B
				;				L'D'E'D E
	EXX			; - - -
	PUSH HL			; save the two exponents.

	LD H,B			; transfer the dividend to H'L'H L
	LD L,C			; 
	EXX			;
	LD H,C			;
	LD L,B			; 

	XOR A			; clear carry bit and accumulator.
	LD B,$DF		; count upwards from -33 decimal
	JR L18B2		; forward to mid-loop entry point DIV-START
; ---
;
;; DIV-LOOP
L18A2:	RLA			; multiply partial quotient by two
	RL C			; setting result bit from carry.
	EXX			;
	RL C			;
	RL B			;
	EXX			;

;; div-34th
L18AB:	ADD HL,HL		;
	EXX			;
	ADC HL,HL		;
	EXX			;
	JR C,L18C2		; forward to SUBN-ONLY

;; DIV-START
L18B2:	SBC HL,DE		; subtract divisor part.
	EXX			;
	SBC HL,DE		;
	EXX			;
	JR NC,L18C9		; forward if subtraction goes to NO-RSTORE

	ADD HL,DE		; else restore	
	EXX			;
	ADC HL,DE		;
	EXX			;
	AND A			; clear carry
	JR L18CA		; forward to COUNT-ONE
; ---
;
;; SUBN-ONLY
L18C2:	AND A			;
	SBC HL,DE		;
	EXX			;
	SBC HL,DE		;
	EXX			;
;
;; NO-RSTORE
L18C9:	SCF			; set carry flag
;
;; COUNT-ONE
L18CA:	INC B			; increment the counter
	JP M,L18A2		; back while still minus to DIV-LOOP

	PUSH AF			;
	JR Z,L18B2		; back to DIV-START
;
; "This jump is made to the wrong place. No 34th bit will ever be obtained
; without first shifting the dividend. Hence important results like 1/10 and
; 1/1000 are not rounded up as they should be. Rounding up never occurs when
; it depends on the 34th bit. The jump should be made to div-34th above."
; - Dr. Frank O'Hara, "The Complete Spectrum ROM Disassembly", 1983,
; published by Melbourne House.
; (Note. on the ZX81 this would be JR Z,L18AB)
;
; However if you make this change, then while (1/2=.5) will now evaluate as
; true, (.25=1/4), which did evaluate as true, no longer does.
;
	LD E,A			;
	LD D,C			;
	EXX			;
	LD E,C			;
	LD D,B			;

	POP AF			;
	RR B			;
	POP AF			;
	RR B			;

	EXX			;
	POP BC			;
	POP HL			;
	LD A,B			;
	SUB C			;
	JP L1810		; jump back to DIVN-EXPT
